from ledsign.device import LEDSign
from ledsign.program import LEDSignProgram
from ledsign.proxy import LEDSignProxyServer
import optparse



_device_list_cache=[]
def _get_device_list() -> list[str]:
	if (not _device_list_cache):
		_device_list_cache.extend(LEDSign.enumerate())
	return _device_list_cache



def _upload_callback(progress:float,is_upload:bool) -> None:
	print(f"{('Upload' if is_upload else 'Clear')} progress: {progress*100:.0f}%")



def main() -> None:
	parser=optparse.OptionParser(prog="ledsign",version="%prog v0.8.1")
	parser.add_option("-d","--device",metavar="DEVICE_PATH|DEVICE_INDEX",dest="device_path",help="open device at DEVICE_PATH, or the device at index DEVICE_INDEX (leave empty to use default device path)")
	parser.add_option("-e","--enumerate",action="store_true",dest="enumerate",help="enumerate all available devices")
	parser.add_option("-x","--enumerate-only",action="store_true",dest="enumerate_only",help="enumerate all available devices and exit (implies --enumerate)")
	parser.add_option("-i","--print-info",action="store_true",dest="print_info",help="print device hardware information")
	parser.add_option("-c","--print-config",action="store_true",dest="print_config",help="print device configuration")
	parser.add_option("-p","--print-driver",action="store_true",dest="print_driver",help="print driver stats")
	parser.add_option("-s","--save",metavar="PROGRAM",dest="save_program",help="save current program into PROGRAM")
	parser.add_option("-u","--upload",metavar="PROGRAM",dest="upload_program",help="upload file PROGRAM to the device (requires read-write mode)")
	parser.add_option("-z","--start-proxy-server",action="store_true",help="start a proxy server (required to communicate with the device)")
	options,args=parser.parse_args()
	if (options.start_proxy_server):
		LEDSignProxyServer()
		input(f"Proxy server running on port {LEDSignProxyServer.PROXY_PORT}, hit 'Enter' to stop...\n")
		return
	device_path=None
	if (not options.enumerate_only):
		if (options.device_path is None):
			if (not _get_device_list()):
				print("No devices found")
				return
			device_path=_get_device_list()[0]
		elif (options.device_path.isnumeric()):
			device_index=int(options.device_path)
			if (device_index>=len(_get_device_list())):
				parser.error("option -d: device index out of range")
				return
			device_path=_get_device_list()[device_index]
		else:
			device_path=options.device_path
	if (options.enumerate or options.enumerate_only):
		devices=LEDSign.enumerate()
		print("system devices:"+"".join([f"\n  {e}" for e in devices]))
		if (options.enumerate_only):
			return
	if (not options.print_info and not options.print_config and not options.print_driver and not options.save_program and not options.upload_program):
		options.print_info=True
		options.print_config=True
		options.print_driver=True
	device=LEDSign.open(device_path)
	try:
		if (options.print_info):
			print(f"device:\n  path: {device.get_path()}\n  storage: {device.get_storage_size()} B\n  hardware: {device.get_hardware().get_string()} ({device.get_hardware().get_user_string()})\n  firmware: {device.get_firmware()}\n  serial number: {device.get_serial_number_str()}")
		if (options.print_config):
			print(f"config:\n  access mode: {device.get_access_mode_str()}\n  power supply: 5V {device.get_psu_current()*1000:.0f}mA ({device.get_psu_current()*5}W)")
		if (options.print_driver):
			print(f"driver:\n  brightness: {device.get_driver_brightness()*100:.0f}%\n  paused: {str(device.is_driver_paused()).lower()}\n  temperature: {device.get_driver_temperature():.1f}*C\n  load: {device.get_driver_load():.1f}%\n  current: {device.get_driver_current_usage()*1000:.0f}mA\n  program time: {device.get_driver_program_time():.3f}s / {device.get_program().get_duration():.3f}s")
		if (options.save_program):
			device.get_program().save(options.save_program)
		if (options.upload_program):
			device.upload_program(LEDSignProgram(device,options.upload_program).compile(),_upload_callback)
	finally:
		device.close()



if (__name__=="__main__"):
	main()



from collections.abc import Callable
from ledsign.checksum import LEDSignCRC
from ledsign.protocol import LEDSignProtocol
import ledsign.program
import struct
import time



__all__=["LEDSignProgramParser","LEDSignCompiledProgram"]



def _bit_permute_step(a,b,c):
	t=((a>>c)^a)&b
	return (a^t)^(t<<c)



class LEDSignProgramParser(object):
	MAX_LINE_EXTRACTION_ERROR=2

	__slots__=["_program","_frame_length","_offset","_stride","_pixel_prev_states","_pixel_curr_states","_pixel_update_stack","_pixel_update_stack_length","_pixel_masks"]

	def __init__(self,program:"LEDSignProgram",frame_length:int,is_compressed:bool) -> None:
		self._program=program
		self._frame_length=frame_length
		self._offset=0
		self._stride=frame_length*12
		self._pixel_prev_states=[0 for _ in range(0,frame_length<<3)]
		self._pixel_curr_states=[0 for _ in range(0,frame_length<<3)]
		self._pixel_update_stack=[0 for _ in range(0,frame_length<<3)]
		self._pixel_update_stack_length=0
		self._pixel_masks=[]
		if (is_compressed):
			for i,pixel in enumerate(program._hardware._pixels):
				if (pixel is None):
					continue
				self._pixel_masks.append(1<<i)
			while (len(self._pixel_masks)<(frame_length<<3)):
				self._pixel_masks.append(0)
		else:
			for i in range(0,frame_length<<3):
				self._pixel_masks.append(1<<i)

	def update(self,data:bytearray) -> None:
		for i in range(0,len(data),12):
			gvec,rvec,bvec=struct.unpack("<III",data[i:i+12])
			rvec=_bit_permute_step(rvec,0x0a0a0a0a,3)
			gvec=_bit_permute_step(gvec,0x0a0a0a0a,3)
			bvec=_bit_permute_step(bvec,0x0a0a0a0a,3)
			rvec=_bit_permute_step(rvec,0x00cc00cc,6)
			gvec=_bit_permute_step(gvec,0x00cc00cc,6)
			bvec=_bit_permute_step(bvec,0x00cc00cc,6)
			rvec=_bit_permute_step(rvec,0x0000f0f0,12)
			gvec=_bit_permute_step(gvec,0x0000f0f0,12)
			bvec=_bit_permute_step(bvec,0x0000f0f0,12)
			rvec=_bit_permute_step(rvec,0x0000ff00,8)
			gvec=_bit_permute_step(gvec,0x0000ff00,8)
			bvec=_bit_permute_step(bvec,0x0000ff00,8)
			j=(self._offset+i)%(self._stride<<1)
			j=j//12+3*self._frame_length*(j>=self._stride)
			for k in range(0,4):
				prev=self._pixel_prev_states[j]
				curr=self._pixel_curr_states[j]
				d=(curr>>24)&0xfffff
				err_r=(rvec&0xff)*d-((curr>>16)&0xff)*(d+1)+((prev>>16)&0xff)
				err_g=(gvec&0xff)*d-((curr>>8)&0xff)*(d+1)+((prev>>8)&0xff)
				err_b=(bvec&0xff)*d-(curr&0xff)*(d+1)+(prev&0xff)
				if (not prev or abs(err_r)+abs(err_g)+abs(err_b)>d*LEDSignProgramParser.MAX_LINE_EXTRACTION_ERROR):
					self._pixel_prev_states[j]=curr;
					if (curr and (not prev or ((curr^prev)&0xffffff))):
						self._pixel_update_stack[self._pixel_update_stack_length]=j
						self._pixel_update_stack_length+=1
					curr&=0xfffff00000000000
				self._pixel_curr_states[j]=(curr&0xffffffffff000000)+((rvec&0xff)<<16)+((gvec&0xff)<<8)+(bvec&0xff)+0x0000100001000000
				rvec>>=8
				gvec>>=8
				bvec>>=8
				j+=self._frame_length
			if ((self._offset+i+12)%(self._stride<<1)):
				continue
			while (self._pixel_update_stack_length):
				value=self._pixel_prev_states[self._pixel_update_stack[0]]
				mask=0
				j=0
				while (j<self._pixel_update_stack_length):
					if (self._pixel_prev_states[self._pixel_update_stack[j]]==value):
						mask|=self._pixel_masks[self._pixel_update_stack[j]]
						self._pixel_update_stack_length-=1
						self._pixel_update_stack[j]=self._pixel_update_stack[self._pixel_update_stack_length]
					else:
						j+=1
				self._program._add_raw_keypoint(value&0xffffff,value>>44,(value>>24)&0xfffff,mask,None)
		self._offset+=len(data)

	def terminate(self) -> None:
		for i in range(0,self._frame_length<<3):
			if ((self._pixel_prev_states[i]^self._pixel_curr_states[i])&0xffffff):
				self._pixel_update_stack[self._pixel_update_stack_length]=i
				self._pixel_update_stack_length+=1
		while (self._pixel_update_stack_length):
			value=self._pixel_curr_states[self._pixel_update_stack[0]]
			mask=0
			i=0
			while (i<self._pixel_update_stack_length):
				if (self._pixel_curr_states[self._pixel_update_stack[i]]==value):
					mask|=self._pixel_masks[self._pixel_update_stack[i]]
					self._pixel_update_stack_length-=1
					self._pixel_update_stack[i]=self._pixel_update_stack[self._pixel_update_stack_length]
				else:
					i+=1
			self._program._add_raw_keypoint(value&0xffffff,value>>44,(value>>24)&0xfffff,mask,None)



class LEDSignCompilationPixel(object):
	__slots__=["r","g","b","prev_r","prev_g","prev_b","mask","kp"]

	def __init__(self,mask:int,kp:"LEDSignKeypoint"):
		self.r=0
		self.g=0
		self.b=0
		self.prev_r=0
		self.prev_g=0
		self.prev_b=0
		self.mask=mask
		self.kp=kp



class LEDSignCompiledProgram(object):
	"""
	Represents a compiled :py:class:`LEDSignProgram` object, returned by :py:func:`LEDSignProgram.compile`.
	"""

	__slots__=["_data","_led_depth","_max_offset","_offset_divisor","_ctrl","_crc"]

	def __init__(self,program:"LEDSignProgram",is_compressed:bool) -> None:
		pixel_states=[]
		if (is_compressed):
			self._led_depth=(program._hardware._pixel_count+7)>>3
			for i,pixel in enumerate(program._hardware._pixels):
				if (pixel is None):
					continue
				pixel_states.append(LEDSignCompilationPixel(1<<i,program._keypoint_list.lookup_increasing(0,1<<i)))
			while (len(pixel_states)<(self._led_depth<<3)):
				pixel_states.append(LEDSignCompilationPixel(0,None))
		else:
			self._led_depth=program._hardware._led_depth
			for i in range(self._led_depth<<3):
				pixel_states.append(LEDSignCompilationPixel(1<<i,program._keypoint_list.lookup_increasing(0,1<<i)))
		self._data=bytearray(program._duration*self._led_depth*24)
		self._max_offset=max(len(self._data)>>2,1)
		self._offset_divisor=max(6*self._led_depth,1)*60
		self._ctrl=(3*self._led_depth)|(len(self._data)<<6)
		for i in range(0,program._duration):
			for j in range(0,self._led_depth<<3):
				pixel=pixel_states[j]
				kp=pixel.kp
				if (kp is None):
					continue
				if (kp.end<=i):
					pixel.r=(kp.rgb>>16)&0xff
					pixel.g=(kp.rgb>>8)&0xff
					pixel.b=kp.rgb&0xff
					pixel.prev_r=(kp.rgb>>16)&0xff
					pixel.prev_g=(kp.rgb>>8)&0xff
					pixel.prev_b=kp.rgb&0xff
					kp=program._keypoint_list.lookup_increasing(kp._key+1,pixel.mask)
					pixel.kp=kp
					if (kp is None):
						continue
				t=max((i-kp.end+1)/kp.duration+1,0)
				pixel.r=round(pixel.prev_r+t*(((kp.rgb>>16)&0xff)-pixel.prev_r))
				pixel.g=round(pixel.prev_g+t*(((kp.rgb>>8)&0xff)-pixel.prev_g))
				pixel.b=round(pixel.prev_b+t*((kp.rgb&0xff)-pixel.prev_b))
			for j in range(0,self._led_depth):
				rveclo=0
				gveclo=0
				bveclo=0
				rvechi=0
				gvechi=0
				bvechi=0
				for k in range(0,4):
					l=k<<3
					pixel=pixel_states[j+k*self._led_depth]
					rveclo|=pixel.r<<l
					gveclo|=pixel.g<<l
					bveclo|=pixel.b<<l
					pixel=pixel_states[j+(k+4)*self._led_depth]
					rvechi|=pixel.r<<l
					gvechi|=pixel.g<<l
					bvechi|=pixel.b<<l
				rveclo=_bit_permute_step(rveclo,0x00aa00aa,7)
				gveclo=_bit_permute_step(gveclo,0x00aa00aa,7)
				bveclo=_bit_permute_step(bveclo,0x00aa00aa,7)
				rvechi=_bit_permute_step(rvechi,0x00aa00aa,7)
				gvechi=_bit_permute_step(gvechi,0x00aa00aa,7)
				bvechi=_bit_permute_step(bvechi,0x00aa00aa,7)
				rveclo=_bit_permute_step(rveclo,0x0000cccc,14)
				gveclo=_bit_permute_step(gveclo,0x0000cccc,14)
				bveclo=_bit_permute_step(bveclo,0x0000cccc,14)
				rvechi=_bit_permute_step(rvechi,0x0000cccc,14)
				gvechi=_bit_permute_step(gvechi,0x0000cccc,14)
				bvechi=_bit_permute_step(bvechi,0x0000cccc,14)
				rveclo=_bit_permute_step(rveclo,0x00f000f0,4)
				gveclo=_bit_permute_step(gveclo,0x00f000f0,4)
				bveclo=_bit_permute_step(bveclo,0x00f000f0,4)
				rvechi=_bit_permute_step(rvechi,0x00f000f0,4)
				gvechi=_bit_permute_step(gvechi,0x00f000f0,4)
				bvechi=_bit_permute_step(bvechi,0x00f000f0,4)
				rveclo=_bit_permute_step(rveclo,0x0000ff00,8)
				gveclo=_bit_permute_step(gveclo,0x0000ff00,8)
				bveclo=_bit_permute_step(bveclo,0x0000ff00,8)
				rvechi=_bit_permute_step(rvechi,0x0000ff00,8)
				gvechi=_bit_permute_step(gvechi,0x0000ff00,8)
				bvechi=_bit_permute_step(bvechi,0x0000ff00,8)
				k=i*self._led_depth*24+j*12
				self._data[k:k+12]=struct.pack("<III",gveclo,rveclo,bveclo)
				k+=12*self._led_depth
				self._data[k:k+12]=struct.pack("<III",gvechi,rvechi,bvechi)
		self._crc=LEDSignCRC(self._data).value

	def __repr__(self) -> str:
		return f"<LEDSignCompiledProgram size={len(self._data)} B>"

	def _upload_to_device(self,device:"LEDSign",callback:Callable[[float,bool],None]|None=None) -> None:
		if (device._hardware._led_depth!=self._led_depth):
			raise ledsign.program.LEDSignProgramError("Mismatched program hardware")
		result=LEDSignProtocol.process_packet(device._handle,LEDSignProtocol.PACKET_TYPE_PROGRAM_CHUNK_REQUEST_DEVICE,LEDSignProtocol.PACKET_TYPE_PROGRAM_SETUP,self._ctrl,self._crc)
		clear_progress_active=True
		prev_clear_progress=0
		if (callback is not None):
			callback(0.0,False)
		while (result[0]!=0xffffffff):
			if (not result[1]):
				if (not clear_progress_active):
					time.sleep(0.02)
			else:
				if (clear_progress_active):
					clear_progress_active=False
				if (callback is not None):
					callback(min(result[0]/len(self._data),1.0),True)
				LEDSignProtocol.process_extended_write(device._handle,self._data[result[0]:result[0]+result[1]])
			result=LEDSignProtocol.process_packet(device._handle,LEDSignProtocol.PACKET_TYPE_PROGRAM_CHUNK_REQUEST_DEVICE,LEDSignProtocol.PACKET_TYPE_PROGRAM_UPLOAD_STATUS)
			if (clear_progress_active and callback is not None and result[2]!=prev_clear_progress):
				prev_clear_progress=result[2]
				callback(min(result[2]/len(self._data),1.0),False)
		if (callback is not None):
			callback(1.0,True)
		device._driver_program_offset_divisor=self._offset_divisor
		device._driver_info_sync_next_time=0
		device._program=None

	def _save_to_file(self,file_path:str) -> None:
		with open(file_path,"wb") as wf:
			wf.write(struct.pack("<II",self._ctrl,self._crc))
			wf.write(self._data)



from collections.abc import Callable
from ledsign.hardware import LEDSignHardware
from ledsign.program import LEDSignProgram
from ledsign.program_io import LEDSignCompiledProgram
from ledsign.protocol import LEDSignProtocol
import time



__all__=["LEDSignDeviceNotFoundError","LEDSignAccessError","LEDSign"]



class LEDSignDeviceNotFoundError(Exception):
	"""
	Raised by :py:func:`LEDSign.open` when no device was found.
	"""



class LEDSignAccessError(Exception):
	"""
	Raised when the device access mode was violated.
	"""



class LEDSign(object):
	"""
	Returned by :py:func:`LEDSign.open`, represents a handle to an LED sign device.

	.. autoattribute:: ACCESS_MODE_NONE
	   :no-value:

	   Access mode representing a device handle without read or write access. At the moment only used by closed device handles.

	.. autoattribute:: ACCESS_MODE_READ
	   :no-value:

	   Access mode representing a device handle with only read permissions.

	.. autoattribute:: ACCESS_MODE_READ_WRITE
	   :no-value:

	   Access mode representing a device handle with both read and write permissions.
	"""

	ACCESS_MODE_NONE:int=0x00
	ACCESS_MODE_READ:int=0x01
	ACCESS_MODE_READ_WRITE:int=0x02

	ACCESS_MODES:dict[int,str]={
		ACCESS_MODE_NONE: "none",
		ACCESS_MODE_READ: "read-only",
		ACCESS_MODE_READ_WRITE: "read-write",
	}

	__slots__=["__weakref__","_path","_handle","_access_mode","_psu_current","_storage_size","_hardware","_firmware","_serial_number","_driver_brightness","_driver_program_paused","_driver_temperature","_driver_load","_driver_program_time","_driver_current_usage","_driver_program_offset_divisor","_driver_info_sync_next_time","_driver_info_sync_interval","_program"]

	def __init__(self,path,handle,config_packet) -> None:
		self._path=path
		self._handle=handle
		self._access_mode=config_packet[6]&0x0f
		self._psu_current=(config_packet[7]&0x7f)/10
		self._storage_size=config_packet[1]<<10
		self._hardware=LEDSignHardware(handle,config_packet[2])
		self._firmware=config_packet[9].hex()
		self._serial_number=config_packet[10]
		self._driver_brightness=config_packet[5]&0x0f
		self._driver_program_paused=not (config_packet[8]&1)
		self._driver_program_offset_divisor=max((config_packet[3]&0xff)<<1,1)*60
		self._driver_info_sync_next_time=0
		self._driver_info_sync_interval=0.5
		self._program=LEDSignProgram._create_unloaded_from_device(self,config_packet[3],config_packet[4])

	def __del__(self) -> None:
		if (self._handle is not None):
			LEDSignProtocol.close(self._handle)

	def __repr__(self) -> str:
		return f"<LEDSign id={self._serial_number:016x} fw={self._firmware}>"

	def _check_if_closed(self) -> None:
		if (self._handle is not None):
			return
		raise LEDSignAccessError("Device handle was closed, no access allowed")

	def _sync_driver_info(self) -> None:
		self._check_if_closed()
		if (time.time()<self._driver_info_sync_next_time):
			return
		driver_status=LEDSignProtocol.process_packet(self._handle,LEDSignProtocol.PACKET_TYPE_LED_DRIVER_STATUS_RESPONSE,LEDSignProtocol.PACKET_TYPE_LED_DRIVER_STATUS_REQUEST)
		self._driver_temperature=437.226612-driver_status[0]*0.468137
		self._driver_load=driver_status[1]/160
		self._driver_program_time=driver_status[2]/self._driver_program_offset_divisor
		self._driver_current_usage=driver_status[3]*1e-6
		self._driver_info_sync_next_time=time.time()+self._driver_info_sync_interval

	def close(self) -> None:
		"""
		Closes the underlying device handle. After a call to this function, all methods except for :py:func:`get_path` and :py:func:`get_access_mode` will raise a :py:exc:`LEDSignAccessError` exception.
		"""
		self._check_if_closed()
		LEDSignProtocol.close(self._handle)
		self._handle=None
		self._path=None
		self._access_mode=LEDSign.ACCESS_MODE_NONE

	def get_path(self) -> str|None:
		"""
		Returns the underlying OS path of the device, or :python:`None` if the device was closed.
		"""
		return self._path

	def get_access_mode(self) -> int:
		"""
		Returns the access mode (permissions) granted by the device. Possible return values are:

		* :py:attr:`ACCESS_MODE_NONE`: No access; device handle was closed
		* :py:attr:`ACCESS_MODE_READ`: Read-only access; program uploads will be rejected
		* :py:attr:`ACCESS_MODE_READ_WRITE`: Full read-write access
		"""
		return self._access_mode

	def get_access_mode_str(self) -> str:
		"""
		Same as :py:func:`get_access_mode`, but returns a stringified versions of the access mode. Possible values are: :python:`"none"`, :python:`"read-only"`, or :python:`"read-write"`.
		"""
		return LEDSign.ACCESS_MODES[self._access_mode]

	def get_psu_current(self) -> float:
		"""
		Returns the configured theoretical current limit of the power supply, in amps. As only 5V power supplies are supported, no explicit voltage getter method is provided.

		.. danger::
		   If the device draws more current than this limit, a device-internal overcurrent safety flag will be raised. Whenever this flag is active, no changes made to the device will be visible.

		   **For safety reasons, this flag can only be cleared from the UI menu, or through device reboots.**
		"""
		self._check_if_closed()
		return self._psu_current

	def get_storage_size(self) -> int:
		"""
		Returns the storage capacity of the device allocated for program storage, in bytes.
		"""
		self._check_if_closed()
		return self._storage_size

	def get_hardware(self) -> LEDSignHardware:
		"""
		Returns a read-only :py:class:`LEDSignHardware` object representing the current external hardware configuration of the device.
		"""
		self._check_if_closed()
		return self._hardware

	def get_firmware(self) -> str:
		"""
		Returns a 40-character hex string uniquely identifying the current firmware versions of the device.
		"""
		self._check_if_closed()
		return self._firmware

	def get_serial_number(self) -> int:
		"""
		Returns the unique 64-bit serial number of the device.
		"""
		self._check_if_closed()
		return self._serial_number

	def get_serial_number_str(self) -> str:
		"""
		Same as :py:func:`get_serial_number`, but formats the serial number as a 16-hex-digit string.
		"""
		self._check_if_closed()
		return f"{self._serial_number:016x}"

	def get_driver_brightness(self) -> float:
		"""
		Returns the current device brightness setting, normalized between :python:`0.0` and :python:`1.0`.

		.. note::
		   If the overcurrent flag was tripped (see :py:func:`get_psu_current` for details), the returned brightness setting may not reflect the real-world conditions.
		"""
		self._check_if_closed()
		return round(self._driver_brightness*20/7)/20

	def is_driver_paused(self) -> bool:
		"""
		Returns :python:`True` if the LED driver is in a paused state (ie. the program playback is frozen), and :python:`False` otherwise.

		.. note::
		   For internal protocol reasons, this flag is **not** periodically synchronized with the device.
		"""
		self._check_if_closed()
		return self._driver_program_paused

	def get_driver_temperature(self) -> float:
		"""
		Returns the current temperature of the LED driver, in degrees Celsius.

		The temperature is fetched periodically from the device, at an interval specified by :py:func:`get_driver_status_reload_time`.
		"""
		self._sync_driver_info()
		return self._driver_temperature

	def get_driver_load(self) -> float:
		"""
		Returns the current driver CPU load, normalized between :python:`0.0` and :python:`1.0`. The returned value represents the frame render duration, measured as a fraction of the overall frame time.

		The driver load is fetched periodically from the device, at an interval specified by :py:func:`get_driver_status_reload_time`.
		"""
		self._sync_driver_info()
		return self._driver_load

	def get_driver_program_time(self) -> float:
		"""
		Returns the current program timestamp. The timestamp wraps around to zero at program duration, ie. :python:`timestamp %= self.get_program().get_duration()`.

		The timestamp is fetched periodically from the device, at an interval specified by :py:func:`get_driver_status_reload_time`.
		"""
		self._sync_driver_info()
		return self._driver_program_time

	def get_driver_current_usage(self) -> float:
		"""
		Returns the current being drawn by the device hardware from the power supply. Internally, this same value is used in the overcurrent protection circuit (see :py:func:`get_psu_current` for more details).

		The current usage is fetched periodically from the device, at an interval specified by :py:func:`get_driver_status_reload_time`.
		"""
		self._sync_driver_info()
		return self._driver_current_usage

	def get_driver_status_reload_time(self) -> float:
		"""
		Returns the current refresh interval (in seconds) used for fetching driver data, or :python:`-1` if caching is disabled. Can be modified using :py:func:`set_driver_status_reload_time`.
		"""
		self._check_if_closed()
		return self._driver_info_sync_interval

	def set_driver_status_reload_time(self,interval:float) -> float:
		"""
		Sets :python:`interval` as the new driver data refresh interval (in seconds), and returns the previous value. Negative values for :python:`interval` disable caching and fetch data whenever it is requested.
		"""
		self._check_if_closed()
		out=self._driver_info_sync_interval
		self._driver_info_sync_interval=(-1 if interval<=0 else interval)
		self._driver_info_sync_next_time+=self._driver_info_sync_interval-out
		return out

	def get_program(self) -> LEDSignProgram:
		"""
		Returns a read-only :py:class:`LEDSignProgram` object containing the current device program. For performance reasons, the program is lazily loaded only when its keypoints are accessed to prevent significant start-up delays.
		"""
		self._check_if_closed()
		return self._program

	def upload_program(self,program:LEDSignCompiledProgram,callback:Callable[[float,bool],None]|None=None) -> None:
		"""
		Uploads the given :python:`program` to the device. If the device was opened in read-only mode, a :py:exc:`LEDSignAccessError` exception will be raised.

		The optional :python:`callback` argument is periodically called throughout the upload process with arguments :python:`(progress, is_first_stage)`.
		"""
		if (not isinstance(program,LEDSignCompiledProgram)):
			raise TypeError(f"Expected 'LEDSignCompiledProgram', got '{program.__class__.__name__}'")
		if (callback is not None and not callable(callback)):
			raise TypeError(f"Expected 'function', got '{callback.__class__.__name__}'")
		if (self._access_mode!=LEDSign.ACCESS_MODE_READ_WRITE):
			raise LEDSignAccessError("Program upload not allowed, Python API configured as read-only")
		program._upload_to_device(self,callback)

	@staticmethod
	def open(path:str|None=None) -> "LEDSign":
		"""
		Opens the device pointed to by :python:`path` (or the default device if no path is provided), and returns its corresponding :py:class:`LEDSign` object. The device path must have been returned by a previous call to :py:func:`enumerate`.

		If the device is not found, a :py:exc:`LEDSignDeviceNotFoundError` will be raised. Additionally, if the device protocol version is incompatible, or if the Python API was disabled in device settings, a :py:exc:`LEDSignUnsupportedProtocolError` or :py:exc:`LEDSignProtocolError` exception will be returned respectively. If a proxy connection cannot be established, or if the proxy server is not running, a :py:exc:`LEDSignProxyError` will be returned (see :ref:`proxy-server` for documentation on how to start the server).

		.. note::
		   Due to the internal USB-level configuration, this library can be used alongside the UI without unintended protocol side effects. This feature allows the UI to be kept open while writing new programs to ease the debugging / visualization process.
		"""
		if (path is None):
			devices=LEDSignProtocol.enumerate()
			if (not devices):
				raise LEDSignDeviceNotFoundError("No device found")
			path=devices[0]
		if (not isinstance(path,str)):
			raise TypeError(f"Expected 'str', got '{path.__class__.__name__}'")
		handle=None
		try:
			handle=LEDSignProtocol.open(path)
			config_packet=LEDSignProtocol.process_packet(handle,LEDSignProtocol.PACKET_TYPE_DEVICE_INFO,LEDSignProtocol.PACKET_TYPE_HOST_INFO,LEDSignProtocol.VERSION)
		except Exception as e:
			if (handle is not None):
				LEDSignProtocol.close(handle)
			raise e
		return LEDSign(path,handle,config_packet)

	@staticmethod
	def enumerate() -> list[str]:
		"""
		Returns a list of available device paths detected by the OS.
		"""
		return LEDSignProtocol.enumerate()



from collections.abc import Iterator



__all__=["LEDSignKeypoint","LEDSignKeypointList"]



class LEDSignKeypoint(object):
	"""
	Represents a single keypoint in a program.

	.. note::
	   For performance reasons, not all generated keypoints are always coalesced together, as this expensive step is only performed during external program loads (ie. programs loaded using :py:func:`LEDSignProgram` or through :py:func:`LEDSign.get_program`).
	"""

	__slots__=["rgb","end","duration","mask","_frame","_key","_subtree_mask","_parent","_color","_nodes"]

	def __init__(self,rgb,end,duration,mask,frame) -> None:
		self.rgb=rgb
		self.end=end
		self.duration=duration
		self.mask=mask
		self._frame=("<unknown>" if frame is None else f"{frame.f_code.co_filename}:{frame.f_lineno}({frame.f_code.co_name})")
		self._key=None
		self._subtree_mask=mask
		self._parent=None
		self._color=0
		self._nodes=[None,None]

	def __repr__(self) -> str:
		return f"<LEDSignKeypoint color=#{self.rgb:06x} duration={self.duration/60:.3f}s end={self.end/60:.3f}s mask={self.mask:x}>"

	def get_rgb(self) -> int:
		"""
		Returns the raw RGB color of the current keypoint (red in MSB, blue in LSB).
		"""
		return self.rgb

	def get_rgb_html(self) -> str:
		"""
		Returns the HTML color code of the current keypoint.
		"""
		return f"#{self.rgb:06x}"

	def get_start(self) -> float:
		"""
		Returns the start time of the keypoint.
		"""
		return (self.end-self.duration)/60

	def get_end(self) -> float:
		"""
		Returns the end time of the keypoint.
		"""
		return self.end/60

	def get_duration(self) -> float:
		"""
		Returns the duration of the keypoint.
		"""
		return self.duration/60

	def get_mask(self) -> int:
		"""
		Returns the current keypoint's pixel mask.
		"""
		return self.mask



class LEDSignKeypointList(object):
	def __init__(self) -> None:
		self.root=None
		self._index=0

	def _rotate_subtree(self,x:LEDSignKeypoint,dir:int) -> None:
		y=x._parent
		z=x._nodes[dir^1]
		x._nodes[dir^1]=z._nodes[dir]
		if (z._nodes[dir]):
			z._nodes[dir]._parent=x
		z._nodes[dir]=x
		x._parent=z
		z._parent=y
		x._subtree_mask=x.mask
		if (x._nodes[0] is not None):
			x._subtree_mask|=x._nodes[0]._subtree_mask
		if (x._nodes[1] is not None):
			x._subtree_mask|=x._nodes[1]._subtree_mask
		z._subtree_mask=z.mask|x._subtree_mask
		if (z._nodes[dir^1] is not None):
			z._subtree_mask|=z._nodes[dir^1]._subtree_mask
		if (y is None):
			self.root=z
		else:
			dir=(x==y._nodes[1])
			y._nodes[dir]=z
			y._subtree_mask=y.mask|z._subtree_mask
			if (y._nodes[dir^1] is not None):
				y._subtree_mask|=y._nodes[dir^1]._subtree_mask

	def clear(self) -> None:
		self.root=None

	def lookup_decreasing(self,key:int,mask:int) -> LEDSignKeypoint|None:
		x=self.root
		while (x is not None and (x._key!=key or not (x.mask&mask))):
			if (key>x._key):
				y=x._nodes[1]
				if (y is not None and (y._subtree_mask&mask)):
					x=y
					continue
				if (x.mask&mask):
					return x
			y=x._nodes[0]
			if (y is not None and (y._subtree_mask&mask)):
				x=y
				continue
			while (True):
				y=x
				x=x._parent
				if (x is None):
					return None
				if (y==x._nodes[1]):
					break
			key=x._key
		return x

	def lookup_increasing(self,key:int,mask:int) -> LEDSignKeypoint|None:
		x=self.root
		while (x is not None and (x._key!=key or not (x.mask&mask))):
			if (key<x._key):
				y=x._nodes[0]
				if (y is not None and (y._subtree_mask&mask)):
					x=y
					continue
				if (x.mask&mask):
					return x
			y=x._nodes[1]
			if (y is not None and (y._subtree_mask&mask)):
				x=y
				continue
			while (True):
				y=x
				x=x._parent
				if (x is None):
					return None
				if (y==x._nodes[0]):
					break
			key=x._key
		return x

	def insert(self,x:LEDSignKeypoint) -> None:
		x._key=(x.end<<44)|self._index
		x._parent=None
		x._nodes=[None,None]
		self._index+=1
		if (self.root is None):
			x._color=0
			self.root=x
			return
		x._color=1
		y=self.root
		while (y._nodes[y._key<x._key] is not None):
			y=y._nodes[y._key<x._key]
		x._parent=y
		y._nodes[y._key<x._key]=x
		while (y is not None and y._color):
			y._subtree_mask=y.mask
			if (y._nodes[0] is not None):
				y._subtree_mask|=y._nodes[0]._subtree_mask
			if (y._nodes[1] is not None):
				y._subtree_mask|=y._nodes[1]._subtree_mask
			z=y._parent
			if (z is None):
				y._color=0
				break
			z._subtree_mask=z.mask
			if (z._nodes[0] is not None):
				z._subtree_mask|=z._nodes[0]._subtree_mask
			if (z._nodes[1] is not None):
				z._subtree_mask|=z._nodes[1]._subtree_mask
			dir=(y==z._nodes[0])
			w=z._nodes[dir]
			if (w is None or not w._color):
				if (x==y._nodes[dir]):
					self._rotate_subtree(y,dir^1)
					y=z._nodes[dir^1]
				self._rotate_subtree(z,dir)
				y._color=0
				z._color=1
				y=z._parent._parent
				break
			y._color=0
			z._color=1
			w._color=0
			x=z
			y=x._parent
		while (y is not None):
			y._subtree_mask=y.mask
			if (y._nodes[0] is not None):
				y._subtree_mask|=y._nodes[0]._subtree_mask
			if (y._nodes[1] is not None):
				y._subtree_mask|=y._nodes[1]._subtree_mask
			y=y._parent

	def iterate(self,mask:int) -> Iterator[LEDSignKeypoint]:
		entry=self.lookup_increasing(0,mask)
		while (entry is not None):
			yield entry
			entry=self.lookup_increasing(entry._key+1,mask)



from ledsign.backend import LEDSignProtocolBackendWindows,LEDSignProtocolBackendLinux
import base64
import hashlib
import socket
import struct
import sys
import threading



__all__=["LEDSignProxyServer","LEDSignBackendProxy","LEDSignProxyError","LEDSignProtocolError"]



class LEDSignProxyError(Exception):
	"""
	Raised when any proxy-related exception is encountered.
	"""



class LEDSignProtocolError(Exception):
	"""
	Raised during any encountered protocol error.
	"""



class LEDSignProxyServerDeviceHandleWrapper(object):
	def __init__(self,id_,file_path,handle):
		self.id_=id_
		self.file_path=file_path
		self.handle=handle
		self.marked=True



class LEDSignProxyServer(object):
	PROXY_PORT=9100

	def __init__(self) -> None:
		self._backend=(LEDSignProtocolBackendWindows if sys.platform=="win32" else LEDSignProtocolBackendLinux)()
		self._server_socket=socket.socket(socket.AF_INET,socket.SOCK_STREAM)
		self._server_socket.setsockopt(socket.SOL_SOCKET,socket.SO_REUSEADDR,1)
		self._server_socket.bind(("127.0.0.1",LEDSignProxyServer.PROXY_PORT))
		self._server_socket.listen(5)
		self._device_handles_by_id={}
		self._device_handles_by_file_path={}
		thr=threading.Thread(target=self._thread)
		thr.daemon=True
		thr.start()

	def _process_packet(self,data):
		if (len(data)<6 or data[0]&1):
			return data[:2]
		if (data[0]==0x00 and len(data)==6):
			return data[:2]+self._process_enumerate_packet()
		if (data[0]==0x02 and len(data)==11):
			return data[:2]+self._process_transfer_ctrl(data)
		if (data[0]==0x04):
			return data[:2]+self._process_transfer_bulk_out(data)
		if (data[0]==0x06):
			return data[:2]+self._process_transfer_bulk_in(data)
		return data[:2]

	def _process_enumerate_packet(self):
		for file_path in self._backend.enumerate():
			if (file_path in self._device_handles_by_file_path):
				self._device_handles_by_file_path[file_path].marked=True
				continue
			handle=self._backend.open(file_path)
			if (handle is None):
				continue
			next_id=1
			while (next_id in self._device_handles_by_id):
				next_id+=1
			wrapper=LEDSignProxyServerDeviceHandleWrapper(next_id,file_path,handle)
			self._device_handles_by_id[next_id]=wrapper
			self._device_handles_by_file_path[file_path]=wrapper
		out=b""
		for wrapper in list(self._device_handles_by_file_path.values()):
			if (not wrapper.marked):
				self._remove_handle(wrapper)
				continue
			wrapper.marked=False
			out+=struct.pack("<I",wrapper.id_)
		return struct.pack("<H",len(self._device_handles_by_file_path))+out

	def _process_transfer_ctrl(self,data):
		type,id_,value,index,length=struct.unpack("<BIHHB",data[1:])
		wrapper=self._device_handles_by_id.get(id_,None)
		if (wrapper is None):
			return b"\x01"
		ret=self._backend.transfer_ctrl(wrapper.handle,type,value,index,length)
		if (ret is None):
			return b"\x01"
		return b"\x00"+ret

	def _process_transfer_bulk_out(self,data):
		endpoint,id_=struct.unpack("<BI",data[1:6])
		wrapper=self._device_handles_by_id.get(id_,None)
		if (wrapper is None):
			return b"\x01"
		return bytearray([not self._backend.transfer_bulk_out(wrapper.handle,endpoint,data[6:])])

	def _process_transfer_bulk_in(self,data):
		endpoint,id_,length=struct.unpack("<BIH",data[1:])
		wrapper=self._device_handles_by_id.get(id_,None)
		if (wrapper is None):
			return b"\x01"
		ret=self._backend.transfer_bulk_in(wrapper.handle,endpoint,length)
		return (b"\x01" if ret is None else b"\x00"+ret)

	def _remove_handle(self,wrapper):
		del self._device_handles_by_id[wrapper.id_]
		del self._device_handles_by_file_path[wrapper.file_path]
		self._backend.close(wrapper.handle)

	def _thread(self):
		while (True):
			cs=self._server_socket.accept()[0]
			cs.settimeout(0.5)
			thr=threading.Thread(target=self._handle_http_connection,args=(cs,),kwargs={})
			thr.daemon=True
			thr.start()

	def _handle_http_connection(self,cs):
		try:
			buffer=cs.recv(16384)
			cs.settimeout(None)
			if (buffer==b"\xffraw\xff"):
				cs.sendall(buffer)
				while (True):
					buffer=cs.recv(16384)
					if (not buffer):
						return
					cs.sendall(self._process_packet(buffer))
				return
			if (not buffer.startswith(b"GET / HTTP/1.1")):
				return
			ws_key=None
			prev_line_was_empty=False
			for line in buffer.split(b"\r\n"):
				if (not line):
					if (prev_line_was_empty):
						break
					prev_line_was_empty=True
					continue
				prev_line_was_empty=False
				line=line.strip().split(b":")
				if (len(line)!=2):
					continue
				key,value=line[0].lower(),line[1].strip()
				if (key==b"sec-websocket-key"):
					ws_key=value
				elif (key==b"connection" and value.lower()!=b"upgrade"):
					return
				elif (key==b"upgrade" and value.lower()!=b"websocket"):
					return
			if (ws_key is None):
				return
			cs.sendall(b"HTTP/1.1 101 Switching Protocols\r\nUpgrade: websocket\r\nConnection: Upgrade\r\nSec-WebSocket-Accept: "+base64.b64encode(hashlib.sha1(ws_key+b"258EAFA5-E914-47DA-95CA-C5AB0DC85B11").digest())+b"\r\n\r\n")
			current_frame_type=-1
			current_frame_data=None
			while (True):
				buffer=cs.recv(16384)
				if (not buffer):
					return
				i=0
				while (i+1<len(buffer)):
					if (buffer[i]&0x70):
						return
					frame_is_last=buffer[i]>>7
					frame_type=buffer[i]&0x0f
					frame_masking_key_present=buffer[i+1]>>7
					frame_length=buffer[i+1]&0x7f
					i+=2
					if (frame_length==126):
						i+=2
						if (i>len(buffer)):
							return
						frame_length=(buffer[i-2]<<8)|buffer[i-1]
					elif (frame_length==127):
						return
					mask=b"\x00\x00\x00\x00"
					if (frame_masking_key_present):
						mask=buffer[i:i+4]
						i+=4
					if (i+frame_length>len(buffer)):
						return
					if (frame_type==0):
						if (current_frame_type==-1):
							return
					elif (frame_type==1 or frame_type==2):
						if (current_frame_type!=-1):
							return
						current_frame_type=frame_type
						current_frame_data=bytearray()
					else:
						return
					j=len(current_frame_data)
					current_frame_data+=buffer[i:i+frame_length]
					for k in range(0,frame_length):
						current_frame_data[j+k]^=mask[k&3]
					i+=frame_length
					if (not frame_is_last):
						continue
					data=self._process_packet(current_frame_data)
					current_frame_type=-1
					current_frame_data=None
					if (len(data)<126):
						packet=bytearray([
							0x82,
							len(data)
						])+data
					else:
						packet=bytearray([
							0x82,
							126,
							len(data)>>8,
							len(data)&0xff
						])+data
					cs.sendall(packet)
		except BrokenPipeError:
			pass
		finally:
			cs.close()




class LEDSignBackendProxy(object):
	def __init__(self) -> None:
		self._client_socket=None

	def enumerate(self) -> list[str]:
		data=self._execute(b"\x00\x00\x00\x00\x00\x00")
		length=struct.unpack("<H",data[2:4])[0]
		return [f"proxy/#{e:08x}" for e in struct.unpack(f"<{length}I",data[4:4+4*length])]

	def open(self,path:str) -> int:
		device_id=int(path[7:],16)
		ret=self._execute(struct.pack("<BBIH2xB",0x02,0x52,device_id,0x5453,5))
		if (ret[2] or len(ret)!=8 or bytes(ret[3:])!=b"reset"):
			raise LEDSignProtocolError("Unable to reset device, Python API disabled")
		return device_id

	def close(self,handle:int) -> None:
		pass

	def io_read_write(self,handle:int,packet:bytes) -> bytearray:
		ret=self._execute(struct.pack("<BBI",0x04,0x04,handle)+packet)
		if (ret[2]):
			raise LEDSignProtocolError("Write to endpoint 04h failed")
		ret=self._execute(struct.pack("<BBIH",0x06,0x04,handle,64))
		if (ret[2] or len(ret)<5):
			raise LEDSignProtocolError("Read from endpoint 84h failed")
		return ret[3:]

	def io_bulk_read(self,handle:int,size:int) -> bytearray:
		ret=self._execute(struct.pack("<BBIH",0x06,0x05,handle,size))
		if (ret[2] or len(ret)!=size+3):
			raise LEDSignProtocolError("Read from endpoint 85h failed")
		return ret[3:]

	def io_bulk_write(self,handle:int,data:bytearray) -> None:
		if (self._execute(struct.pack("<BBI",0x04,0x05,handle)+data)[2]):
			raise LEDSignProtocolError("Write to endpoint 04h failed")

	def _execute(self,data):
		if (self._client_socket is None):
			self._client_socket=socket.socket(socket.AF_INET,socket.SOCK_STREAM)
			try:
				self._client_socket.connect(("127.0.0.1",LEDSignProxyServer.PROXY_PORT))
			except ConnectionRefusedError:
				raise LEDSignProxyError("Unable to connect to proxy server, run 'ledsign -z' to start one")
			self._client_socket.sendall(b"\xffraw\xff")
			self._client_socket.recv(5)
		self._client_socket.sendall(data)
		return self._client_socket.recv(16384)



if (__name__=="__main__"):
	LEDSignProxyServer()
	input()



import ctypes
import os
import struct



__all__=["LEDSignProtocolBackendWindows","LEDSignProtocolBackendLinux"]



class LEDSignProtocolBackendWindows(object):
	CM_GET_DEVICE_INTERFACE_LIST_ALL_DEVICES=0x00000000
	CR_BUFFER_SMALL=0x0000001a
	GENERIC_WRITE=0x40000000
	GENERIC_READ=0x80000000
	FILE_SHARE_READ=0x00000001
	FILE_SHARE_WRITE=0x00000002
	OPEN_EXISTING=0x00000003
	FILE_ATTRIBUTE_NORMAL=0x00000080
	FILE_FLAG_OVERLAPPED=0x40000000

	def __init__(self) -> None:
		import ctypes.wintypes
		GUID=type("GUID",(ctypes.Structure,),{"_fields_":[("Data1",ctypes.wintypes.DWORD),("Data2",ctypes.wintypes.WORD),("Data3",ctypes.wintypes.WORD),("Data4",ctypes.wintypes.BYTE*8)]})
		PGUID=ctypes.POINTER(GUID)
		self.CM_Get_Device_Interface_List_SizeA=ctypes.windll.cfgmgr32.CM_Get_Device_Interface_List_SizeA
		self.CM_Get_Device_Interface_List_SizeA.argtypes=(ctypes.wintypes.PULONG,PGUID,ctypes.wintypes.LPVOID,ctypes.wintypes.ULONG)
		self.CM_Get_Device_Interface_List_SizeA.restype=ctypes.wintypes.DWORD
		self.CM_Get_Device_Interface_ListA=ctypes.windll.cfgmgr32.CM_Get_Device_Interface_ListA
		self.CM_Get_Device_Interface_ListA.argtypes=(PGUID,ctypes.wintypes.LPVOID,ctypes.wintypes.PCHAR,ctypes.wintypes.ULONG,ctypes.wintypes.ULONG)
		self.CM_Get_Device_Interface_ListA.restype=ctypes.wintypes.DWORD
		self.CreateFileA=ctypes.windll.kernel32.CreateFileA
		self.CreateFileA.argtypes=(ctypes.wintypes.LPCSTR,ctypes.wintypes.DWORD,ctypes.wintypes.DWORD,ctypes.wintypes.LPVOID,ctypes.wintypes.DWORD,ctypes.wintypes.DWORD,ctypes.wintypes.HANDLE)
		self.CreateFileA.restype=ctypes.wintypes.HANDLE
		self.CloseHandle=ctypes.windll.kernel32.CloseHandle
		self.CloseHandle.argtypes=(ctypes.wintypes.HANDLE,)
		self.CloseHandle.restype=ctypes.wintypes.BOOL
		self.WinUsb_Initialize=ctypes.windll.winusb.WinUsb_Initialize
		self.WinUsb_Initialize.argtypes=(ctypes.wintypes.HANDLE,ctypes.POINTER(ctypes.wintypes.HANDLE))
		self.WinUsb_Initialize.restype=ctypes.wintypes.BOOL
		self.WinUsb_Free=ctypes.windll.winusb.WinUsb_Free
		self.WinUsb_Free.argtypes=(ctypes.wintypes.HANDLE,)
		self.WinUsb_Free.restype=ctypes.wintypes.BOOL
		self.WinUsb_ControlTransfer=ctypes.windll.winusb.WinUsb_ControlTransfer
		self.WinUsb_ControlTransfer.argtypes=(ctypes.wintypes.HANDLE,ctypes.c_ulonglong,ctypes.wintypes.PCHAR,ctypes.wintypes.ULONG,ctypes.wintypes.PULONG,ctypes.wintypes.LPVOID)
		self.WinUsb_ControlTransfer.restype=ctypes.wintypes.BOOL
		self.WinUsb_ReadPipe=ctypes.windll.winusb.WinUsb_ReadPipe
		self.WinUsb_ReadPipe.argtypes=(ctypes.wintypes.HANDLE,ctypes.wintypes.CHAR,ctypes.wintypes.PCHAR,ctypes.wintypes.ULONG,ctypes.wintypes.PULONG,ctypes.wintypes.LPVOID)
		self.WinUsb_ReadPipe.restype=ctypes.wintypes.BOOL
		self.WinUsb_WritePipe=ctypes.windll.winusb.WinUsb_WritePipe
		self.WinUsb_WritePipe.argtypes=(ctypes.wintypes.HANDLE,ctypes.wintypes.CHAR,ctypes.wintypes.PCHAR,ctypes.wintypes.ULONG,ctypes.wintypes.PULONG,ctypes.wintypes.LPVOID)
		self.WinUsb_WritePipe.restype=ctypes.wintypes.BOOL
		self.WinUsb_GetAssociatedInterface=ctypes.windll.winusb.WinUsb_GetAssociatedInterface
		self.WinUsb_GetAssociatedInterface.argtypes=(ctypes.wintypes.HANDLE,ctypes.wintypes.CHAR,ctypes.POINTER(ctypes.wintypes.HANDLE))
		self.WinUsb_GetAssociatedInterface.restype=ctypes.wintypes.BOOL
		self.winusb_registry_guid_ref=ctypes.byref(GUID.from_buffer(bytearray(b"\x58\x30\xf5\xfc\x7b\x99\x21\x4b\xaf\xd6\xe5\x65\x04\x39\x23\xc1")))
		self.GetLastError=ctypes.windll.kernel32.GetLastError
		self.GetLastError.argtypes=tuple()
		self.GetLastError.restype=ctypes.wintypes.DWORD

	def enumerate(self) -> list[str]:
		while (True):
			length=ctypes.wintypes.ULONG(0)
			if (self.CM_Get_Device_Interface_List_SizeA(ctypes.byref(length),self.winusb_registry_guid_ref,0,LEDSignProtocolBackendWindows.CM_GET_DEVICE_INTERFACE_LIST_ALL_DEVICES)):
				raise OSError("CM_Get_Device_Interface_List_SizeA error")
			data=(ctypes.wintypes.CHAR*length.value)()
			ret=self.CM_Get_Device_Interface_ListA(self.winusb_registry_guid_ref,0,data,length,LEDSignProtocolBackendWindows.CM_GET_DEVICE_INTERFACE_LIST_ALL_DEVICES)
			if (not ret):
				return [e for e in bytes(data).decode("utf-8").split("\x00") if e]
			if (ret==LEDSignProtocolBackendWindows.CR_BUFFER_SMALL):
				continue
			raise OSError("CM_Get_Device_Interface_ListA error")

	def open(self,path:str) -> tuple[int,int,int]:
		handle=self.CreateFileA(path.encode("utf-8"),LEDSignProtocolBackendWindows.GENERIC_WRITE|LEDSignProtocolBackendWindows.GENERIC_READ,LEDSignProtocolBackendWindows.FILE_SHARE_READ|LEDSignProtocolBackendWindows.FILE_SHARE_WRITE,0,LEDSignProtocolBackendWindows.OPEN_EXISTING,LEDSignProtocolBackendWindows.FILE_ATTRIBUTE_NORMAL|LEDSignProtocolBackendWindows.FILE_FLAG_OVERLAPPED,0)
		if (handle==0xffffffffffffffff):
			raise OSError("Device already in use")
		winusb_handle=ctypes.wintypes.HANDLE(0)
		if (not self.WinUsb_Initialize(handle,ctypes.byref(winusb_handle))):
			self.CloseHandle(handle)
			raise OSError("Device already in use")
		interface_handle=ctypes.wintypes.HANDLE(0)
		if (not self.WinUsb_GetAssociatedInterface(winusb_handle,0,ctypes.byref(interface_handle))):
			self.WinUsb_Free(winusb_handle.value)
			self.CloseHandle(handle)
			raise OSError("Device error")
		return (handle,winusb_handle.value,interface_handle.value)

	def close(self,handles:tuple[int,int,int]) -> None:
		self.WinUsb_Free(handles[2])
		self.WinUsb_Free(handles[1])
		self.CloseHandle(handles[0])

	def transfer_ctrl(self,handles:tuple[int,int,int],type:int,value:int,index:int,length:int) -> None|bytearray:
		buffer=(ctypes.wintypes.CHAR*length)()
		transferred=ctypes.c_ulong(0)
		if (not self.WinUsb_ControlTransfer(handles[1],0xc0|(type<<8)|(value<<16)|(index<<32)|(length<<48),buffer,length,ctypes.byref(transferred),0)):
			return None
		return bytearray(buffer)[:length]

	def transfer_bulk_out(self,handles:tuple[int,int,int],endpoint:int,data:bytes) -> bool:
		data=bytearray(data)
		transferred=ctypes.c_ulong(0)
		return self.WinUsb_WritePipe(handles[1+(endpoint>3)],endpoint&0x7f,(ctypes.c_char*len(data)).from_buffer(data),len(data),ctypes.byref(transferred),0) and transferred.value==len(data)

	def transfer_bulk_in(self,handles:tuple[int,int,int],endpoint:int,length:int) -> None|bytearray:
		out=(ctypes.c_char*length)()
		transferred=ctypes.c_ulong(0)
		return (None if not self.WinUsb_ReadPipe(handles[1+(endpoint>3)],(endpoint&0x7f)|0x80,out,length,ctypes.byref(transferred),0) else bytearray(out)[:transferred.value])



class LEDSignProtocolBackendLinux(object):
	USBDEVFS_BULK=0xc0185502
	USBDEVFS_CLAIMINTERFACE=0x8004550f
	USBDEVFS_CONTROL=0xc0185500

	def __init__(self) -> None:
		import ctypes.util
		self.ioctl=ctypes.CDLL(ctypes.util.find_library("c"),use_errno=True).ioctl
		self.ioctl.argtypes=(ctypes.c_int,ctypes.c_ulong,ctypes.c_char_p)
		self.ioctl.restype=ctypes.c_int

	def enumerate(self) -> list[str]:
		out=[]
		for name in os.listdir("/sys/bus/usb/devices"):
			if (not os.path.exists(f"/sys/bus/usb/devices/{name}/idVendor")):
				continue
			with open(f"/sys/bus/usb/devices/{name}/idVendor","rb") as rf:
				if (rf.read()!=b"fff0\n"):
					continue
			with open(f"/sys/bus/usb/devices/{name}/idProduct","rb") as rf:
				if (rf.read()!=b"1000\n"):
					continue
			with open(f"/sys/bus/usb/devices/{name}/busnum","rb") as rf:
				busnum=int(rf.read())
			with open(f"/sys/bus/usb/devices/{name}/devnum","rb") as rf:
				devnum=int(rf.read())
			out.append(f"/dev/bus/usb/{busnum:03d}/{devnum:03d}")
		return out

	def open(self,path:str) -> int:
		handle=os.open(path,os.O_RDWR)
		return (None if self.ioctl(handle,LEDSignProtocolBackendLinux.USBDEVFS_CLAIMINTERFACE,struct.pack("<I",0))<0 or self.ioctl(handle,LEDSignProtocolBackendLinux.USBDEVFS_CLAIMINTERFACE,struct.pack("<I",1))<0 else handle)

	def close(self,handle:int) -> None:
		os.close(handle)

	def transfer_ctrl(self,handle:int,type:int,value:int,index:int,length:int) -> None|bytearray:
		buffer=(ctypes.c_uint8*length)()
		if (self.ioctl(handle,LEDSignProtocolBackendLinux.USBDEVFS_CONTROL,struct.pack("<BBHHHI4xQ",0xc0,type,value,index,length,3000,ctypes.addressof(buffer)))!=length):
			return None
		return bytearray(buffer)[:length]

	def transfer_bulk_out(self,handle:int,endpoint:int,data:bytes) -> bool:
		data=bytearray(data)
		return self.ioctl(handle,LEDSignProtocolBackendLinux.USBDEVFS_BULK,struct.pack("<III4xQ",endpoint&0x7f,len(data),3000,ctypes.addressof((ctypes.c_uint8*len(data)).from_buffer(data))))==len(data)

	def transfer_bulk_in(self,handle:int,endpoint:int,length:int) -> None|bytearray:
		out=(ctypes.c_uint8*length)()
		ret=self.ioctl(handle,LEDSignProtocolBackendLinux.USBDEVFS_BULK,struct.pack("<III4xQ",(endpoint&0x7f)|0x80,length,3000,ctypes.addressof(out)))
		return (None if ret<0 else bytearray(out)[:ret])



import setuptools



VERSION="0.8.1"



with open("README.md","r") as rf:
	description=rf.read()
setuptools.setup(
	name="ledsign",
	version=VERSION,
	description="LED Sign Python API",
	long_description=description,
	long_description_content_type="text/markdown",
	author="Krzesimir Hyżyk",
	url="https://github.com/krzem5/ledsign",
	packages=["ledsign"],
    python_requires=">=3.9",
	classifiers=[
		"Development Status :: 4 - Beta",
		"Intended Audience :: End Users/Desktop",
		"Natural Language :: English",
		"Operating System :: Microsoft :: Windows",
		"Operating System :: POSIX :: Linux",
		"Programming Language :: Python",
		"Programming Language :: Python :: 3.9",
		"Programming Language :: Python :: 3.10",
		"Programming Language :: Python :: 3.11",
		"Programming Language :: Python :: 3.12",
		"Programming Language :: Python :: 3.13"
	],
	license="BSD-3-Clause",
	install_requires=[],
	project_urls={
		"Source Code": "https://github.com/krzem5/ledsign",
		"Documentation": "https://ledsign.readthedocs.io/en/latest/",
	}
)
